/*
  */

#include "structure.h"

static CanonImage *cim;

static inline char get(int x, int y) { return cim->d[y * cim->w + x]; }
static inline char get_p(Point p) { return get(p.x,p.y); }

#define START_MAIN {200,200}
#define MIN_COLUMNS         6
#define INTERESTING_COLUMNS 7
#define TEXT_COLUMNS        2
#define MAX_COLUMNS         7

static Rect mainr = { START_MAIN,START_MAIN };
static int commbasey, comminty;
static int colrightx[INTERESTING_COLUMNS];
static int text_h=-1, columns=-1;
static OcrReader *rd;

const CanonColourInfo canoncolourinfos[]= {
  { 0x475A5E, '*' }, /* edge */
  { 0x2C5F7A, '*' }, /* edge just under box heading shadow */
  { 0x7D9094, '+' }, /* interbox */

  { 0xBDC5BF, ' ' }, /* background - pale  Sugar cane, etc. */
  { 0xADB5AF, ' ' }, /* background - dark                   */
  { 0xC7E1C3, ' ' }, /* background - pale  Swill, etc.      */
  { 0xB5CFB1, ' ' }, /* background - dark                   */
  { 0xD6CEB0, ' ' }, /* background - pale  Madder, etc.     */
  { 0xC8C0A2, ' ' }, /* background - dark                   */
  { 0xE0E1D3, ' ' }, /* background - pale  Lorandite, etc.  */
  { 0xD0D1C3, ' ' }, /* background - dark                   */
  { 0xE5E6C1, ' ' }, /* background - pale  Cloth            */
  { 0xD7D8B3, ' ' }, /* background - dark                   */
  { 0xEDDED9, ' ' }, /* background - pale  Dye              */
  { 0xDACBC6, ' ' }, /* background - dark                   */
  { 0xD3DEDF, ' ' }, /* background - pale  Paint            */
  { 0xC5D0D1, ' ' }, /* background - dark                   */
  { 0xDCD1CF, ' ' }, /* background - pale  Enamel           */
  { 0xCEC3C1, ' ' }, /* background - dark                   */
  { 0xF3F6F5, ' ' }, /* background - pale  fruit            */
  { 0xE2E7E5, ' ' }, /* background - dark                   */

  { 0x000000, 'o' }, /* foreground */
  { 0xD4B356, ' ' }, /* background (cursor) */
  { 0xFFFFFF, 'o' }, /* foreground (cursor) */

  { 0x5B93BF, '_' }, /* selector dropdown background */
  { 0xD7C94F, 'X' }, /* selector dropdown foreground */
  { 0,0 }
};


static void mustfail1(const char *file, int line, const char *what) {
  fprintf(stderr,
	  "\n\n"
	  "Unable to figure out contents YPP client display.\n"
	  " Check that your client is logged in has the correct display.\n"
	  " If that isn't the problem, please report this as a fault.\n\n"
	  "Technical details:"
	  " %s:%d: requirement failed: %s\n",
	  file, line, what);
}
static void mustfail2(void) NORET;
static void mustfail2(void) {
  fprintf(stderr, "\n\nGiving up.\n");
  exit(8);
}

#define MUST(x, ifnot) do{			\
    if (!(x)) {					\
      mustfail1(__FILE__,__LINE__,#x);		\
      ifnot;					\
      mustfail2();				\
    }						\
  }while(0)

#define MP(v) fprintf(stderr," %s=%d,%d",#v,(v).x,(v).y)
#define MI(v) fprintf(stderr," %s=%d",   #v,(v))
#define MC(v) fprintf(stderr," %s='%c'", #v,(v))
#define MS(v) fprintf(stderr," %s=\"%s\"", #v,(v))
#define MSB(v) fprintf(stderr," %s", (v))
#define MR(v) fprintf(stderr," %s=%d,%d..%d,%d",\
                      #v,(v).tl.x,(v).tl.y,(v).br.x,(v).br.y)


static void require_rectangle(int tlx, int tly, int brx, int bry,
			      const char *ok) {
  Point p;
  for (p.x=tlx; p.x<=brx; p.x++)
    for (p.y=tly; p.y<=bry; p.y++) {
      int c= get_p(p);
      MUST( strchr(ok,c), ({
	     Rect rm={{tlx,tly},{brx,bry}};
	     MR(rm);MP(p);MS(ok);
      }));
    }
}
static void require_rectangle_r(Rect rr, const char *ok) {
  require_rectangle(rr.tl.x,rr.tl.y, rr.br.x,rr.br.y, ok);
}

static void debug_rect(const char *what, int whati, Rect rr) {
  if (!DEBUGP(rect)) return;
  int y,w;
  fprintf(debug, "%s %d: %d,%d..%d,%d:\n", what, whati,
	  rr.tl.x,rr.tl.y, rr.br.x,rr.br.y);
  w= rr.br.x - rr.tl.x + 1;
  for (y=rr.tl.y; y<=rr.br.y; y++) {
    fprintf(debug, "%4d%*s|", y, rr.tl.x,"");
    fwrite(cim->d + y*cim->w + rr.tl.x, 1, w, debug);
    fputc('|',debug);
    fputc('\n',debug);
  }
  debug_flush();
}

#define WALK_UNTIL(point,coord,increm,last,edge)			\
  for (;;) {								\
    if ((point).coord == (last)+(increm)) break;			\
    if (get_p((point)) == (edge)) { (point).coord -= (increm); break; }	\
    (point).coord += (increm);						\
  }

#define WALK_UNTIL_MUST(point,coord,increm,last,edge)	\
  do {							\
    WALK_UNTIL(point,coord,increm,last,edge);		\
    MUST( (point).coord != (last)+(increm),		\
	  MP(point); MI(increm); MI(last); MC(edge);	\
	  );						\
  }while(0)

void find_structure(CanonImage *im) {
  cim= im;
  
  Rect whole = { {0,0}, {cim->w-1,cim->h-1} };

  WALK_UNTIL_MUST(mainr.tl, x,-1, whole.tl.x, '*');
  WALK_UNTIL_MUST(mainr.tl, y,-1, whole.tl.y, '*');
  WALK_UNTIL_MUST(mainr.br, x,+1, whole.br.x, '*');
  WALK_UNTIL_MUST(mainr.br, y,+1, whole.br.y, '*');

  require_rectangle(mainr.tl.x-1, mainr.tl.y, mainr.tl.x-1, mainr.br.y, "*");
  require_rectangle(mainr.br.x+1, mainr.tl.y, mainr.br.x+1, mainr.br.y, "*");
  require_rectangle(mainr.tl.x, mainr.tl.y-1, mainr.br.x, mainr.tl.y-1, "*");
  require_rectangle(mainr.tl.x, mainr.br.y+1, mainr.br.x, mainr.br.y+1, "*");

#define CHECK_STRIP_BORDER(tlbr,xy,increm)	\
  do {						\
    Point csb_p;				\
    Rect csb_r;					\
    csb_p= mainr.tl;				\
    csb_p.xy= mainr.tlbr.xy;			\
    if (get_p(csb_p)=='+') {			\
      csb_r= mainr;				\
      csb_r.tl.xy= csb_p.xy;			\
      csb_r.br.xy= csb_p.xy;			\
      require_rectangle_r(csb_r, "+");		\
      mainr.tlbr.xy += increm;			\
    }						\
  } while(0)

  debug_rect("mainr",0, mainr);

  CHECK_STRIP_BORDER(tl,x,+1);
  CHECK_STRIP_BORDER(tl,y,+1);
  CHECK_STRIP_BORDER(br,x,-1);
  CHECK_STRIP_BORDER(br,y,-1);

  debug_rect("mainr",1, mainr);

  Point up = START_MAIN;
  WALK_UNTIL_MUST(up, y,-1, mainr.tl.y, '+');

  Point down = START_MAIN;
  down.y++;
  WALK_UNTIL_MUST(down, y,+1, mainr.br.y, '+');

  if (DEBUGP(rect)) {
    int xscaleunit, y,x;
    for (y=0, xscaleunit=1; y<4; y++, xscaleunit*=10) {
      fprintf(debug,"     ");
      for (x=0; x<=cim->w; x++) {
	if (x % xscaleunit) fputc(' ',debug);
	else fprintf(debug,"%d",(x / xscaleunit)%10);
      }
      fputc('\n',debug);
    }
  }

  commbasey= up.y;
  comminty= down.y - up.y + 2;

  Point across= { mainr.tl.x, commbasey };
  int colno=0;
  for (;;) {
    MUST( get_p(across) != '+', MI(colno);MP(across);MR(mainr);MI(commbasey) );
    WALK_UNTIL(across, x,+1, mainr.br.x, '+');
    MUST( colno < MAX_COLUMNS, MP(across);MR(mainr);MI(commbasey); );
    int colrx= across.x;
    if (colrx > mainr.br.x) colrx= mainr.br.x;
    if (colno < INTERESTING_COLUMNS)
      colrightx[colno]= colrx;
      
    colno++;
    
    if (across.x >= mainr.br.x-1)
      break;

    across.x++;
    require_rectangle(across.x,mainr.tl.y, across.x,mainr.br.y, "+");
    across.x++;
  }
  MUST( colno >= MIN_COLUMNS, MI(colno);MR(mainr);MP(across); );

#define SET_ONCE(var,val) do{						\
    int v= (val);							\
    if ((var)==-1) (var)= v;						\
    else MUST( (var) == v, MSB(#var);MI((var));MI(v);MR(mainr); );	\
  }while(0)

  SET_ONCE(columns, colno);
  SET_ONCE(text_h, comminty - 1);
}		    

CanonImage *alloc_canon_image(int w, int h) {
  CanonImage *im= mmalloc(sizeof(CanonImage) + w*h);
  im->w= w;
  im->h= h;
  memset(im->d,'?',w*h);
  return im;
}

static void file_read_image_ppm(FILE *f) {
  struct pam inpam;
  unsigned char rgb_buf[3];
  CanonImage *im;

  pnm_readpaminit(f, &inpam, sizeof(inpam));
  if (!(inpam.maxval == 255 &&
	inpam.bytes_per_sample == 1 &&
	inpam.format == RPPM_FORMAT))
    fatal("PNM screenshot(s) file must be 8bpp 1 byte per sample RGB");

  CANONICALISE_IMAGE(im, inpam.width, inpam.height, {
    int r= fread(&rgb_buf,1,3,f);
    sysassert(!ferror(f));
    if (r!=3) fatal("PNM screenshot(s) file ends unexpectedly");

    rgb=
	((unsigned long)rgb_buf[0]<<16) |
	((unsigned long)rgb_buf[1]<<8) |
	               (rgb_buf[2]);
  });

  sysassert(!ferror(screenshots_file));

  if (!(npages < MAX_PAGES))
    fatal("Too many images in screenshots file; max is %d.\n", MAX_PAGES);

  page_images[npages++]= im;
}

void read_one_screenshot(void) {
  progress("reading screenshot...");
  file_read_image_ppm(screenshots_file);
  progress_log("read screenshot.");
}

void read_screenshots(void) {
  struct stat stab;
  
  sysassert(! fstat(fileno(screenshots_file), &stab) );
  
  for (;;) {
    if (S_ISREG(stab.st_mode)) {
      long pos= ftell(screenshots_file);
      if (pos == stab.st_size) break;
    } else {
      int c= fgetc(screenshots_file);
      if (c==EOF) break;
      ungetc(c, screenshots_file);
    }
    progress("reading screenshot %d...",npages);
    file_read_image_ppm(screenshots_file);
  }
  sysassert(!ferror(screenshots_file));
  progress_log("read %d screenshots.",npages);
}

static void find_commodity(int offset, Rect *rr) {
  /* rr->tl.x==-1 if offset out of range */
  rr->tl.y= commbasey - offset*comminty;
  rr->br.y= rr->tl.y + comminty-2;
  if (rr->tl.y < mainr.tl.y || rr->br.y > mainr.br.y) { rr->tl.x=-1; return; }
  
  rr->tl.x= mainr.tl.x;
  rr->br.x= mainr.br.x;

  if (rr->tl.y > mainr.tl.y)
    require_rectangle(rr->tl.x,rr->tl.y-1, rr->br.x,rr->tl.y-1, "+");
  if (rr->br.y < mainr.tl.y)
    require_rectangle(rr->tl.x,rr->br.y+1, rr->br.x,rr->br.y+1, "+");
}

static void find_table_entry(Rect commod, int colno, Rect *cellr) {
  cellr->tl.y= commod.tl.y;
  cellr->br.y= commod.br.y;
  cellr->tl.x= !colno ? commod.tl.x : colrightx[colno-1]+2;
  cellr->br.x=                        colrightx[colno];
  debug_rect("cell", colno, *cellr);
  require_rectangle_r(*cellr, " o");
}

static void ocr_rectangle(Rect r, const OcrCellType ct) {
  OcrResultGlyph *results, *res;

  int w= r.br.x - r.tl.x + 1;
  Pixcol cols[w+1];
  int x,y;
  for (x=0; x<w; x++) {
    Pixcol cx, rv;
    for (y=0, cx=0, rv=1; y<text_h; y++, rv<<=1) {
      Point here= { x+r.tl.x, y+r.tl.y };
      int pixel= get_p(here);
      switch (pixel) {
      case ' ':           break;
      case 'o': cx |= rv; break;
      default:
	MUST(!"wrong pixel",
	     MC(pixel);MP(here);MSB(ocr_celltype_name(ct));MR(r); );
      }
    }
    cols[x]= cx;
  }
  cols[w]= 0;

  results= ocr(rd,ct,w,cols);
  for (res=results; res->s; res++)
    printf("%s",res->s);
}

void analyse(void) {
  Rect thisr, entryr;
  int page, tryrect, colno;

  for (page=0; page<npages; page++) {
    find_structure(page_images[page]);

    if (!rd)
      rd= ocr_init(text_h);

    for (tryrect= +cim->h; tryrect >= -cim->h; tryrect--) {
      find_commodity(tryrect, &thisr);
      if (thisr.tl.x < 0)
	continue;
      debug_rect("commod",tryrect, thisr);

      const char *tab= "";
      for (colno=0; colno<columns; colno++) {
	find_table_entry(thisr,colno,&entryr);
	fputs(tab, stdout);
	ocr_rectangle(entryr,
		      colno<TEXT_COLUMNS
		      ? &ocr_celltype_text
		      : &ocr_celltype_number);
	tab= "\t";
      }
      fputs("\n", stdout);
      sysassert(!ferror(stdout));
      sysassert(!fflush(stdout));
    }
  }
}
