#ifndef STRUCTURE_H
#define STRUCTURE_H


#include "convert.h"


typedef struct {
  unsigned long rgb; /* on screen */
  char c; /* canonical */
} CanonColourInfo;

extern const CanonColourInfo canoncolourinfos[];

CanonImage *alloc_canon_image(int w, int h);

#define CANONICALISE_IMAGE(im,w,h, COMPUTE_RGB) do{		\
    /* compute_rgb should be a number of statements, or		\
     * a block, which assigns to				\
     *   unsigned long rgb;					\
     * given the values of					\
     *   int x,y;						\
     * all of which are anamorphic.  Result is stored in im.	\
     * The COMPUTE_RGB is executed exactly once for		\
     * each pixel in reading order.				\
     */								\
    (im)= alloc_canon_image((w), (h));				\
								\
    int x,y;							\
    for (y=0; y<(h); y++) {					\
      for (x=0; x<(w); x++) {					\
        const CanonColourInfo *cci;				\
        unsigned long rgb;					\
	COMPUTE_RGB;						\
	for (cci=canoncolourinfos; cci->c; cci++) {		\
	  if (cci->rgb == rgb) {				\
	    (im)->d[y*(w) + x]= cci->c;				\
	    break;						\
	  }							\
	}							\
      }								\
      if (DEBUGP(rect)) {					\
	fprintf(debug, "%4d ",y);				\
	fwrite(im->d + y*w, 1,w, debug);			\
	fputc('\n',debug);					\
      }								\
    }								\
    debug_flush();						\
  }while(0)


#endif /*STRUCTURE_H*/
